/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_fcp.c 1.370 2004/09/20 20:04:55EDT sf_support Exp  $
 */

#include <linux/version.h>
#include <linux/config.h>
#include <linux/init.h>
#include <linux/blkdev.h>
#include <linux/ctype.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/pci.h>
#include <linux/smp_lock.h>
#include <linux/spinlock.h>
#include <linux/timer.h>
#include <linux/utsname.h>

#include <asm/byteorder.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
#include <scsi/scsi_transport_fc.h>
#endif

#include "lpfc_sli.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
#include "lpfc.h"
#include "lpfc_fcp.h"
#include "lpfc_hw.h"
#include "lpfc_logmsg.h"
#include "lpfc_mem.h"
#include "lpfc_version.h"
#include "lpfc_compat.h"
#include "lpfc_crtn.h"

static char *lpfc_drvr_name = LPFC_DRIVER_NAME;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
static struct scsi_transport_template *lpfc_transport_template = NULL;
#endif

static struct list_head lpfc_hba_list = LIST_HEAD_INIT(lpfc_hba_list);

static const char *
lpfc_info(struct Scsi_Host *host)
{
	struct lpfc_hba    *phba = (struct lpfc_hba *) host->hostdata[0];
	int len;
	static char  lpfcinfobuf[128];
	
	memset(lpfcinfobuf,0,128);
	if (phba && phba->pcidev){
	        lpfc_get_hba_model_desc(phba, 0, lpfcinfobuf);
		len = strlen(lpfcinfobuf);	
		snprintf(lpfcinfobuf + len,
			128-len,
	       		" on PCI bus %02x device %02x irq %d",
			phba->pcidev->bus->number,
		 	phba->pcidev->devfn,
			phba->pcidev->irq);
	}
	return lpfcinfobuf;
}

static void
lpfc_jedec_to_ascii(int incr, char hdw[])
{
	int i, j;
	for (i = 0; i < 8; i++) {
		j = (incr & 0xf);
		if (j <= 9)
			hdw[7 - i] = 0x30 +  j;
		 else
			hdw[7 - i] = 0x61 + j - 10;
		incr = (incr >> 4);
	}
	hdw[8] = 0;
	return;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
static ssize_t
lpfc_drvr_version_show(struct class_device *cdev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, LPFC_MODULE_DESC "\n");
}

static ssize_t
lpfc_info_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	return snprintf(buf, PAGE_SIZE, "%s\n",lpfc_info(host));
}

static ssize_t
lpfc_serialnum_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	return snprintf(buf, PAGE_SIZE, "%s\n",phba->SerialNumber);
}

static ssize_t
lpfc_fwrev_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	char fwrev[32];
	lpfc_decode_firmware_rev(phba, fwrev, 1);
	return snprintf(buf, PAGE_SIZE, "%s\n",fwrev);
}

static ssize_t
lpfc_hdw_show(struct class_device *cdev, char *buf)
{
	char hdw[9];
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	lpfc_vpd_t *vp = &phba->vpd;
	lpfc_jedec_to_ascii(vp->rev.biuRev, hdw);
	return snprintf(buf, PAGE_SIZE, "%s\n", hdw);	
}
static ssize_t
lpfc_option_rom_version_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	return snprintf(buf, PAGE_SIZE, "%s\n", phba->OptionROMVersion);
}
static ssize_t
lpfc_num_discovered_ports_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	return snprintf(buf, PAGE_SIZE, "%d\n", phba->fc_map_cnt +
							phba->fc_unmap_cnt);
}
static ssize_t
lpfc_node_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	uint64_t node_name = 0;
	memcpy (&node_name, &phba->fc_nodename, sizeof (struct lpfc_name));
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", be64_to_cpu(node_name));
}
static ssize_t
lpfc_port_name_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	uint64_t port_name = 0;
	memcpy (&port_name, &phba->fc_portname, sizeof (struct lpfc_name));
	return snprintf(buf, PAGE_SIZE, "0x%llx\n", be64_to_cpu(port_name));
}
static ssize_t
lpfc_did_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	return snprintf(buf, PAGE_SIZE, "0x%x\n", phba->fc_myDID);
}
static ssize_t
lpfc_events_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	int i = 0, len = 0, get = phba->hba_event_put;
	struct lpfc_hba_event *rec;

	if (get == phba->hba_event_get)
		 return snprintf(buf, PAGE_SIZE, "None\n");
	
	for (i = 0; i < MAX_HBAEVT; i++) {
		if (get == 0)
			get = MAX_HBAEVT;
		get--;
		rec = &phba->hbaevt[get];
		switch (rec->fc_eventcode) {
			case 0:
				len += snprintf(buf+len, PAGE_SIZE-len,
						"---------");
				break;
			case HBA_EVENT_RSCN:
				len += snprintf(buf+len, PAGE_SIZE-len,
						"RSCN     ");
				break;
			case HBA_EVENT_LINK_UP:
				len += snprintf(buf+len, PAGE_SIZE-len,
						 "LINK UP  ");
				break;
			case HBA_EVENT_LINK_DOWN:
				len += snprintf(buf+len, PAGE_SIZE-len,
							"LINK DOWN");
				break;
			default:
				len += snprintf(buf+len, PAGE_SIZE-len,
						"?????????");
				break;

		}
		len += snprintf(buf+len, PAGE_SIZE-len, " %d,%d,%d,%d\n",
				 rec->fc_evdata1, rec->fc_evdata2,
				 rec->fc_evdata3, rec->fc_evdata4);
	}
	return len;
}
static ssize_t
lpfc_nport_evt_cnt_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	return snprintf(buf, PAGE_SIZE, "%d\n", phba->nport_event_cnt);
}

static ssize_t
lpfc_board_online_show(struct class_device *cdev, char *buf)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];

	if (!phba) return 0;

	if (phba->fc_flag & FC_OFFLINE_MODE)
		return snprintf(buf, PAGE_SIZE, "0\n");
	else
		return snprintf(buf, PAGE_SIZE, "1\n");
}

static ssize_t
lpfc_board_online_store(struct class_device *cdev, const char *buf,
								size_t count)
{
	struct Scsi_Host *host = class_to_shost(cdev);
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
 	int val=0;

	if (!phba) return 0;

 	if (sscanf(buf, "%d", &val) != 1)
		return 0;
	
	if (val && (phba->fc_flag & FC_OFFLINE_MODE)) {
		lpfc_online(phba);
	}
	else if (!val && !(phba->fc_flag & FC_OFFLINE_MODE)) {
		lpfc_offline(phba);
	}

	return strlen(buf);
}

#define lpfc_param_show(attr)	\
static ssize_t \
lpfc_##attr##_show(struct class_device *cdev, char *buf) \
{ \
 	struct Scsi_Host *host = class_to_shost(cdev);\
 	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];\
 	int val = 0;\
 	if (phba){\
 		val = phba->cfg_##attr;\
 		return snprintf(buf, PAGE_SIZE, "%d\n",\
 				phba->cfg_##attr);\
 	}\
 	return 0;\
}

#define lpfc_param_store(attr, minval, maxval)	\
static ssize_t \
lpfc_##attr##_store(struct class_device *cdev, const char *buf, size_t count) \
{ \
 	struct Scsi_Host *host = class_to_shost(cdev);\
 	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];\
 	int val=0;\
 	if (sscanf(buf, "%d", &val) != 1)\
 		return 0;\
 	if (phba){\
 		if (val >= minval && val <= maxval) {\
 			phba->cfg_##attr = val;\
 			return strlen(buf);\
 		}\
 	}\
 	return 0;\
}

#define LPFC_ATTR_R_NOINIT(name, desc) \
extern int lpfc_##name;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc);\
lpfc_param_show(name)\
static CLASS_DEVICE_ATTR(lpfc_##name, S_IRUGO , lpfc_##name##_show, NULL)

#define LPFC_ATTR_R(name, defval, minval, maxval, desc) \
static int lpfc_##name = defval;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc);\
lpfc_param_show(name)\
static CLASS_DEVICE_ATTR(lpfc_##name, S_IRUGO , lpfc_##name##_show, NULL)

#define LPFC_ATTR_RW(name, defval, minval, maxval, desc) \
static int lpfc_##name = defval;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc);\
lpfc_param_show(name)\
lpfc_param_store(name, minval, maxval)\
static CLASS_DEVICE_ATTR(lpfc_##name, S_IRUGO | S_IWUSR,\
			 lpfc_##name##_show, lpfc_##name##_store)

static CLASS_DEVICE_ATTR(info, S_IRUGO, lpfc_info_show, NULL);
static CLASS_DEVICE_ATTR(serialnum, S_IRUGO, lpfc_serialnum_show, NULL);
static CLASS_DEVICE_ATTR(fwrev, S_IRUGO, lpfc_fwrev_show, NULL);
static CLASS_DEVICE_ATTR(hdw, S_IRUGO, lpfc_hdw_show, NULL);
static CLASS_DEVICE_ATTR(option_rom_version, S_IRUGO,
					lpfc_option_rom_version_show, NULL);
static CLASS_DEVICE_ATTR(num_discovered_ports, S_IRUGO,
					lpfc_num_discovered_ports_show, NULL);
static CLASS_DEVICE_ATTR(node_name, S_IRUGO, lpfc_node_name_show, NULL);
static CLASS_DEVICE_ATTR(port_name, S_IRUGO, lpfc_port_name_show, NULL);
static CLASS_DEVICE_ATTR(portfcid, S_IRUGO, lpfc_did_show, NULL);
static CLASS_DEVICE_ATTR(events, S_IRUGO, lpfc_events_show, NULL);
static CLASS_DEVICE_ATTR(nport_evt_cnt, S_IRUGO, lpfc_nport_evt_cnt_show, NULL);
static CLASS_DEVICE_ATTR(lpfc_drvr_version, S_IRUGO, lpfc_drvr_version_show,
			 NULL);
static CLASS_DEVICE_ATTR(board_online, S_IRUGO | S_IWUSR,
			 lpfc_board_online_show, lpfc_board_online_store);
#else
#define LPFC_ATTR_R_NOINIT(name, desc) \
extern int lpfc_##name;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc);\

#define LPFC_ATTR_R(name, defval, minval, maxval, desc) \
static int lpfc_##name = defval;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc)

#define LPFC_ATTR_RW(name, defval, minval, maxval, desc) \
static int lpfc_##name = defval;\
module_param(lpfc_##name, int, 0);\
MODULE_PARM_DESC(lpfc_##name, desc)
#endif

/*
# lpfc_log_verbose: Only turn this flag on if you are willing to risk being
# deluged with LOTS of information.
# You can set a bit mask to record specific types of verbose messages:
#
# LOG_ELS                       0x1        ELS events
# LOG_DISCOVERY                 0x2        Link discovery events
# LOG_MBOX                      0x4        Mailbox events
# LOG_INIT                      0x8        Initialization events
# LOG_LINK_EVENT                0x10       Link events
# LOG_IP                        0x20       IP traffic history
# LOG_FCP                       0x40       FCP traffic history
# LOG_NODE                      0x80       Node table events
# LOG_MISC                      0x400      Miscellaneous events
# LOG_SLI                       0x800      SLI events
# LOG_CHK_COND                  0x1000     FCP Check condition flag
# LOG_LIBDFC                    0x2000     LIBDFC events
# LOG_ALL_MSG                   0xffff     LOG all messages
*/
LPFC_ATTR_RW(log_verbose, 0x0, 0x0, 0xffff, "Verbose logging bit-mask");

/*
# lun_queue_depth:  This parameter is used to limit the number of outstanding
# commands per FCP LUN. Value range is [1,128]. Default value is 30.
*/
LPFC_ATTR_R(lun_queue_depth, 30, 1, 128,
	    "Max number of FCP commands we can queue to a specific LUN");

/*
# Some disk devices have a "select ID" or "select Target" capability.
# From a protocol standpoint "select ID" usually means select the
# Fibre channel "ALPA".  In the FC-AL Profile there is an "informative
# annex" which contains a table that maps a "select ID" (a number
# between 0 and 7F) to an ALPA.  By default, for compatibility with
# older drivers, the lpfc driver scans this table from low ALPA to high
# ALPA.
#
# Turning on the scan-down variable (on  = 1, off = 0) will
# cause the lpfc driver to use an inverted table, effectively
# scanning ALPAs from high to low. Value range is [0,1]. Default value is 1.
#
# (Note: This "select ID" functionality is a LOOP ONLY characteristic
# and will not work across a fabric. Also this parameter will take
# effect only in the case when ALPA map is not available.)
*/
LPFC_ATTR_RW(scan_down, 1, 0, 1,
	     "Start scanning for devices from highest ALPA to lowest");

/*
# lpfc_nodev_tmo: If set, it will hold all I/O errors on devices that disappear
# until the timer expires. Value range is [0,255]. Default value is 20.
# NOTE: this MUST be less then the SCSI Layer command timeout - 1.
*/
LPFC_ATTR_RW(nodev_tmo, 20, 0, 255,
	     "Seconds driver will hold I/O waiting for a device to come back");

/*
# lpfc_topology:  link topology for init link
#            0x0  = attempt loop mode then point-to-point
#            0x02 = attempt point-to-point mode only
#            0x04 = attempt loop mode only
#            0x06 = attempt point-to-point mode then loop
# Set point-to-point mode if you want to run as an N_Port.
# Set loop mode if you want to run as an NL_Port. Value range is [0,0x6].
# Default value is 0.
*/
LPFC_ATTR_R(topology, 0, 0, 6, "Select Fibre Channel topology");

/*
# lpfc_link_speed: Link speed selection for initializing the Fibre Channel
# connection.
#       0 = auto select (default)
#       1 = 1 Gigabaud
#       2 = 2 Gigabaud
#       4 = 10 Gigabaud
#       8 = 4 Gigabaud
# Value range is [0,8]. Default value is 0.
*/
LPFC_ATTR_R(link_speed, 0, 0, 8, "Select link speed");

/*
# lpfc_fcp_class:  Determines FC class to use for the FCP protocol.
# Value range is [2,3]. Default value is 3.
*/
LPFC_ATTR_R(fcp_class, 3, 2, 3,
	     "Select Fibre Channel class of service for FCP sequences");

/*
# lpfc_use_adisc: Use ADISC for FCP rediscovery instead of PLOGI. Value range
# is [0,1]. Default value is 0.
*/
LPFC_ATTR_RW(use_adisc, 0, 0, 1,
	     "Use ADISC on rediscovery to authenticate FCP devices");

/*
# lpfc_ack0: Use ACK0, instead of ACK1 for class 2 acknowledgement. Value
# range is [0,1]. Default value is 0.
*/
LPFC_ATTR_R(ack0, 0, 0, 1, "Enable ACK0 support");

/*
# If automap is set, SCSI IDs for all FCP nodes without
# consistent bindings will be automatically generated.
# If new FCP devices are added to the network when the system is down,
# there is no guarantee that these SCSI IDs will remain the same
# when the system is booted again.
# The bind method of the port is used as the binding method of
# automap devices to preserve SCSI IDs between link down and link up.
# If automap is 0, only devices with consistent bindings will be
# recognized by the system. User can change the automap property
# of port instance X by changing the value of lpfcX_automap parameter.
# Value range is [0,1]. Default value is 1.
*/
LPFC_ATTR_RW(automap, 1, 0, 1,
	    "Automatically bind FCP devices as they are discovered");

/*
# lpfc_fcp_bind_method: It specifies the method of binding to be used for each
# port. This  binding method is used for consistent binding and automaped
# binding. A value of 1 will force WWNN binding, value of 2 will force WWPN
# binding, value of 3 will force DID binding and value of 4 will force the
# driver to derive binding from ALPA. Any consistent binding whose type does
# not match with the bind method of the port will be ignored. Value range
# is [1,4]. Default value is 2.
*/
LPFC_ATTR_RW(fcp_bind_method, 2, 0, 4,
	    "Select the bind method to be used");

/*
# lpfc_cr_delay & lpfc_cr_count: Default values for I/O colaesing
# cr_delay (msec) or cr_count outstanding commands. cr_delay can take
# value [0,63]. cr_count can take value [0,255]. Default value of cr_delay
# is 0. Default value of cr_count is 0. The cr_count feature is disabled if
# cr_delay is set to 0.
*/
static int lpfc_cr_delay = 0;
module_param(lpfc_cr_delay, int , 0);
MODULE_PARM_DESC(lpfc_cr_delay, "A count of milliseconds after which an"
		"interrupt response is generated");

static int lpfc_cr_count = 1;
module_param(lpfc_cr_count, int, 0);
MODULE_PARM_DESC(lpfc_cr_count, "A count of I/O completions after which an"
		"interrupt response is generated");

/*
# lpfc_fdmi_on: controls FDMI support.
#       0 = no FDMI support
#       1 = support FDMI without attribute of hostname
#       2 = support FDMI with attribute of hostname
# Value range [0,2]. Default value is 0.
*/
LPFC_ATTR_RW(fdmi_on, 0, 0, 2, "Enable FDMI support");

/*
# Specifies the maximum number of ELS cmds we can have outstanding (for
# discovery). Value range is [1,64]. Default value = 1.
*/
static int lpfc_discovery_threads = 1;
module_param(lpfc_discovery_threads, int, 0);
MODULE_PARM_DESC(lpfc_discovery_threads, "Maximum number of ELS commands"
		 "during discovery");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
static ssize_t
sysfs_stats_read(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	struct Scsi_Host *host = dev_to_shost(container_of(kobj, struct device,
									kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	portstatistics_t hs;
	LPFC_MBOXQ_t *pmboxq;
	MAILBOX_t *pmb;
	int rc = 0;
	struct lpfc_sli *psli = &phba->sli;

	pmboxq = mempool_alloc(phba->mbox_mem_pool, GFP_ATOMIC);
	if (!pmboxq)
		return -ENOMEM;
	memset(pmboxq, 0, sizeof (LPFC_MBOXQ_t));

	pmb = &pmboxq->mb;
	pmb->mbxCommand = MBX_READ_STATUS;
	pmb->mbxOwner = OWN_HOST;
	pmboxq->context1 = NULL;

	if ((phba->fc_flag & FC_OFFLINE_MODE) ||
	    (!(psli->sliinit.sli_flag & LPFC_SLI2_ACTIVE))){
		rc = lpfc_sli_issue_mbox(phba, pmboxq, MBX_POLL);
	} else
		rc = lpfc_sli_issue_mbox_wait(phba, pmboxq, phba->fc_ratov * 2);

	if (rc != MBX_SUCCESS) {
		if (pmboxq) {
			if (rc == MBX_TIMEOUT)
				pmboxq->mbox_cmpl = 0;
			else
				mempool_free( pmboxq, phba->mbox_mem_pool);
		}
		return -ENODEV;
	}

	hs.TxFrames = pmb->un.varRdStatus.xmitFrameCnt;
	hs.RxFrames = pmb->un.varRdStatus.rcvFrameCnt;
	hs.TxWords = (pmb->un.varRdStatus.xmitByteCnt * 256);
	hs.RxWords = (pmb->un.varRdStatus.rcvByteCnt * 256);

	memset((void *)pmboxq, 0, sizeof (LPFC_MBOXQ_t));
	pmb->mbxCommand = MBX_READ_LNK_STAT;
	pmb->mbxOwner = OWN_HOST;
	pmboxq->context1 = NULL;

	if ((phba->fc_flag & FC_OFFLINE_MODE) ||
	    (!(psli->sliinit.sli_flag & LPFC_SLI2_ACTIVE)))
		rc = lpfc_sli_issue_mbox(phba, pmboxq, MBX_POLL);
	else
		rc = lpfc_sli_issue_mbox_wait(phba, pmboxq, phba->fc_ratov * 2);

	if (rc != MBX_SUCCESS) {
		if (pmboxq) {
			if (rc == MBX_TIMEOUT)
				pmboxq->mbox_cmpl = 0;
			else
				mempool_free( pmboxq, phba->mbox_mem_pool);
		}
		return -ENODEV;
	}

	hs.LinkFailureCount = pmb->un.varRdLnk.linkFailureCnt;
	hs.LossOfSyncCount = pmb->un.varRdLnk.lossSyncCnt;
	hs.LossOfSignalCount = pmb->un.varRdLnk.lossSignalCnt;
	hs.PrimitiveSeqProtocolErrCount = pmb->un.varRdLnk.primSeqErrCnt;
	hs.InvalidTxWordCount = pmb->un.varRdLnk.invalidXmitWord;
	hs.InvalidCRCCount = pmb->un.varRdLnk.crcCnt;
	hs.ErrorFrames = pmb->un.varRdLnk.crcCnt;

	if (phba->fc_topology == TOPOLOGY_LOOP) {
		hs.LIPCount = (phba->fc_eventTag >> 1);
		hs.NOSCount = -1;
	} else {
		hs.LIPCount = -1;
		hs.NOSCount = (phba->fc_eventTag >> 1);
	}

	hs.DumpedFrames = -1;
/* FIX ME */
	/*hs->SecondsSinceLastReset = (jiffies - lpfc_loadtime) / HZ;*/

	memcpy(buf, &hs, sizeof(hs));

	mempool_free(pmboxq, phba->mbox_mem_pool);

	return (rc);
}

static struct bin_attribute sysfs_stats_attr = {
	.attr = {
		.name = "stats",
		.mode = S_IRUGO,
		.owner = THIS_MODULE,
	},
	.size = 0,
	.read = sysfs_stats_read,
};

static int
dfc_rsp_data_copy(struct lpfc_hba * phba,  uint8_t * outdataptr,
					 DMABUFEXT_t * mlist, uint32_t size)
{
	DMABUFEXT_t *mlast = NULL;
	int cnt, offset = 0;
	struct list_head head, *curr, *next;

	if (!mlist) /* FIX ME - fix the return values */
		return 0;

	list_add_tail(&head, &mlist->dma.list);

	list_for_each_safe(curr, next, &head) {
		mlast = list_entry(curr, DMABUFEXT_t , dma.list);
		if (!size)
			break;

		/* We copy chunks of 4K */
		cnt = size > 4096 ? 4096: size;

		if (outdataptr) {
			pci_dma_sync_single_for_device(phba->pcidev,
			    mlast->dma.phys, LPFC_BPL_SIZE, PCI_DMA_TODEVICE);
	
			 memcpy(outdataptr + offset,
						mlast->dma.virt, cnt);
		}
		offset += cnt;
		size -= cnt;
	}
	list_del(&head);
	return 0;
}

static int
dfc_cmd_data_free(struct lpfc_hba * phba, DMABUFEXT_t * mlist)
{
	DMABUFEXT_t *mlast;
	struct pci_dev *pcidev = phba->pcidev;
	struct list_head head, *curr, *next;

	if (!mlist) /* FIX ME - need different return value */
		return 0;

	list_add_tail(&head, &mlist->dma.list);

	list_for_each_safe(curr, next, &head) {
		mlast = list_entry(curr, DMABUFEXT_t , dma.list);
		if (mlast->dma.virt) {
			dma_free_coherent(&pcidev->dev, 
					  mlast->size, 
					  mlast->dma.virt, 
					  mlast->dma.phys);

		}
		kfree(mlast);
	}
	return 0;
}

static DMABUFEXT_t *
dfc_cmd_data_alloc(struct lpfc_hba * phba, char *indataptr,
					 struct ulp_bde64 * bpl, uint32_t size)
{
	DMABUFEXT_t *mlist = NULL, *dmp;
	int cnt, offset = 0, i = 0, rc = 0;
	struct pci_dev *pcidev = phba->pcidev;

	while (size) {
		cnt = size > 4096 ? 4096: size;

		dmp = kmalloc(sizeof (DMABUFEXT_t), GFP_KERNEL);
		if (!dmp)
			goto dfc_cmd_data_alloc_exit;

		INIT_LIST_HEAD(&dmp->dma.list);

		if (mlist)
			list_add_tail(&dmp->dma.list, &mlist->dma.list);
		else
			mlist = dmp;

		dmp->dma.virt = dma_alloc_coherent(&pcidev->dev, 
						   cnt, 
						   &(dmp->dma.phys), 
						   GFP_KERNEL);

		if (!dmp->dma.virt) /* FIX ME - who free's the list ?*/
			goto dfc_cmd_data_alloc_free_dmp;

		dmp->size = cnt;

		if (!indataptr) {
			bpl->tus.f.bdeFlags = BUFF_USE_RCV;
		} else {
	
			/*rc = copy_from_user(dmp->dma.virt, indataptr + offset,
									cnt);*/
			memcpy(dmp->dma.virt, indataptr+offset, cnt);
			if (rc) /* FIX ME - who free's the list ?*/
				goto dfc_cmd_data_alloc_free_dmp;
			bpl->tus.f.bdeFlags = 0;

			pci_dma_sync_single_for_device(phba->pcidev,
				dmp->dma.phys, LPFC_BPL_SIZE, PCI_DMA_TODEVICE);
		}

		/* build buffer ptr list for IOCB */
		bpl->addrLow = le32_to_cpu( putPaddrLow(dmp->dma.phys) );
		bpl->addrHigh = le32_to_cpu( putPaddrHigh(dmp->dma.phys) );
		bpl->tus.f.bdeSize = (ushort) cnt;
		bpl->tus.w = le32_to_cpu(bpl->tus.w);
		bpl++;

		i++;
		offset += cnt;
		size -= cnt;
	}

	mlist->flag = i;
	return (mlist);
dfc_cmd_data_alloc_free_dmp:
	kfree(dmp);
dfc_cmd_data_alloc_exit:
	dfc_cmd_data_free(phba, mlist);
	return (0);
}

static ssize_t
sysfs_ctpass_write(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
						struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	struct lpfc_nodelist *pndl;
	struct ulp_bde64 *bpl;
	struct lpfc_iocbq *cmdiocbq = 0, *rspiocbq = 0;
	DMABUFEXT_t *indmp = 0, *outdmp = 0;
	IOCB_t *cmd = 0, *rsp = 0;
	struct lpfc_dmabuf *bmp = 0;
	struct lpfc_sli *psli = &phba->sli;
	struct lpfc_sli_ring *pring = &psli->ring[LPFC_ELS_RING];
	int i, rc = -4;
	int reqbfrcnt, snsbfrcnt;
	uint32_t timeout;
	unsigned long iflag;
	uint32_t portid;      /* Port to send this to. */
	typedef struct tagctpassthruinput { 
		uint32_t portid;
		uint32_t reqsize;
		uint32_t rspsize;
	} ctpassthruinput_t;

	spin_lock_irqsave(phba->host->host_lock, iflag);

	reqbfrcnt = ((ctpassthruinput_t*)buf)->reqsize;
	snsbfrcnt = ((ctpassthruinput_t*)buf)->rspsize;
	portid = ((ctpassthruinput_t*)buf)->portid;

	if((portid & CT_DID_MASK) != CT_DID_MASK)
		goto ctpassthru_exit;

	if (!reqbfrcnt || !snsbfrcnt ||
		(reqbfrcnt > PAGE_SIZE - sizeof(ctpassthruinput_t)) ||
		(snsbfrcnt > PAGE_SIZE)) {
		rc = -ERANGE;
		goto ctpassthru_exit;
	}

	pndl = lpfc_findnode_did(phba, NLP_SEARCH_MAPPED | NLP_SEARCH_UNMAPPED,
								 portid);
	if(!pndl || pndl->nlp_flag & NLP_ELS_SND_MASK) {
		rc = -ENODEV;
		goto ctpassthru_exit;
	}

	if (!(psli->sliinit.sli_flag & LPFC_SLI2_ACTIVE)) {
		rc = -EACCES;
		goto ctpassthru_exit;
	}

	cmdiocbq = mempool_alloc(phba->iocb_mem_pool, GFP_ATOMIC);
	if (!cmdiocbq) {
		rc = -ENOMEM;
		goto ctpassthru_exit;
	}
	memset(cmdiocbq, 0, sizeof (struct lpfc_iocbq));
	cmd = &cmdiocbq->iocb;

	rspiocbq = mempool_alloc(phba->iocb_mem_pool, GFP_ATOMIC);
	if (!rspiocbq) {
		rc = -ENOMEM;
		goto ctpassthru_freecmdiocbq;
	}
	memset(rspiocbq, 0, sizeof (struct lpfc_iocbq));
	rsp = &rspiocbq->iocb;

	bmp = kmalloc(sizeof (struct lpfc_dmabuf), GFP_ATOMIC);
	if (!bmp) {
		rc = -ENOMEM;
		goto ctpassthru_freerspiocbq;
	}

	bmp->virt = lpfc_mbuf_alloc(phba, 0, &bmp->phys);
	if (!bmp->virt) {
		rc = -ENOMEM;
		goto ctpassthru_freebmp;
	}

	INIT_LIST_HEAD(&bmp->list);
	bpl = (struct ulp_bde64 *) bmp->virt;
	spin_unlock_irqrestore(phba->host->host_lock, iflag);
	indmp = dfc_cmd_data_alloc(phba, buf + sizeof(ctpassthruinput_t), bpl,
								reqbfrcnt);
	spin_lock_irqsave(phba->host->host_lock, iflag);
	if (!indmp) {
		rc = -ENOMEM;
		goto ctpassthru_freembuf;
	}

	bpl += indmp->flag; /* flag contains total number of BPLs for xmit */

	spin_unlock_irqrestore(phba->host->host_lock, iflag);
	outdmp = dfc_cmd_data_alloc(phba, 0, bpl, snsbfrcnt);
	spin_lock_irqsave(phba->host->host_lock, iflag);
	if (!outdmp) {
		rc = -ENOMEM;
		goto ctpassthru_free_indmp;
	}
	outdmp->data = snsbfrcnt;
	outdmp->uniqueid = current_thread_info()->task->pid;
	
	cmd->un.genreq64.bdl.ulpIoTag32 = 0;
	cmd->un.genreq64.bdl.addrHigh = putPaddrHigh(bmp->phys);
	cmd->un.genreq64.bdl.addrLow = putPaddrLow(bmp->phys);
	cmd->un.genreq64.bdl.bdeFlags = BUFF_TYPE_BDL;
	cmd->un.genreq64.bdl.bdeSize =
	    (outdmp->flag + indmp->flag) * sizeof (struct ulp_bde64);
	cmd->ulpCommand = CMD_GEN_REQUEST64_CR;
	cmd->un.genreq64.w5.hcsw.Fctl = (SI | LA);
	cmd->un.genreq64.w5.hcsw.Dfctl = 0;
	cmd->un.genreq64.w5.hcsw.Rctl = FC_UNSOL_CTL;
	cmd->un.genreq64.w5.hcsw.Type = FC_COMMON_TRANSPORT_ULP;
	cmd->ulpIoTag = lpfc_sli_next_iotag(phba, pring);
	cmd->ulpTimeout = 5;
	cmd->ulpBdeCount = 1;
	cmd->ulpLe = 1;
	cmd->ulpClass = CLASS3;
	cmd->ulpContext = pndl->nlp_rpi;
	cmd->ulpOwner = OWN_CHIP;
	cmdiocbq->context1 = NULL;
	cmdiocbq->context2 = NULL;
	cmdiocbq->iocb_flag |= LPFC_IO_LIBDFC;

	if (cmd->ulpTimeout < (phba->fc_ratov * 2 + LPFC_DRVR_TIMEOUT))
		timeout = phba->fc_ratov * 2 + LPFC_DRVR_TIMEOUT;
	else
		timeout = cmd->ulpTimeout;

	for (rc = -1, i = 0; i < 4 && rc != IOCB_SUCCESS; i++) {
		spin_unlock_irqrestore(phba->host->host_lock, iflag);
		rc = lpfc_sli_issue_iocb_wait(phba, pring, cmdiocbq,
					     SLI_IOCB_USE_TXQ, rspiocbq,
					     timeout);
		spin_lock_irqsave(phba->host->host_lock, iflag);
	}

	if (rc != IOCB_SUCCESS) {
		rc = -EACCES;
		goto ctpassthru_free_outdmp;
	}

	if (!rsp->ulpStatus) {
		outdmp->flag = rsp->un.genreq64.bdl.bdeSize;
	} else {
		if (rsp->ulpStatus == IOSTAT_LOCAL_REJECT) {
			switch (rsp->un.ulpWord[4] & 0xff) {
			case IOERR_SEQUENCE_TIMEOUT:
				rc = -ETIMEDOUT;
				break;
			case IOERR_INVALID_RPI:
				rc = -EFAULT;
				break;
			default:
				rc = -EACCES;
				break;
			}
			goto ctpassthru_free_outdmp;
		}
	}
	if (outdmp->flag > snsbfrcnt) {
		rc = -ERANGE; /* C_CT Request error */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_LIBDFC,
			       "%d:1208 C_CT Request error Data: x%x x%x\n",
				phba->brd_no,
			       outdmp->flag, 4096);
		goto ctpassthru_free_outdmp;
	}
	list_add(&outdmp->list, &phba->ctrspbuflist);
	rc = reqbfrcnt; 
	goto ctpassthru_free_indmp;

ctpassthru_free_outdmp:
	dfc_cmd_data_free(phba, outdmp);
ctpassthru_free_indmp:
	dfc_cmd_data_free(phba, indmp);
ctpassthru_freembuf:
	lpfc_mbuf_free(phba, bmp->virt, bmp->phys);
ctpassthru_freebmp:
	kfree(bmp);
ctpassthru_freerspiocbq:
	mempool_free(rspiocbq, phba->iocb_mem_pool);
ctpassthru_freecmdiocbq:
	mempool_free(cmdiocbq, phba->iocb_mem_pool);
ctpassthru_exit:
	spin_unlock_irqrestore(phba->host->host_lock, iflag); /* remove */
	return rc;
}

static ssize_t
sysfs_ctpass_read(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	int rc = -EIO, uniqueid;
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
						struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	DMABUFEXT_t *outdmp = 0, *tmpoutdmp;
	
	uniqueid = current_thread_info()->task->pid;

	list_for_each_entry_safe(outdmp, tmpoutdmp, &phba->ctrspbuflist, list){
		if (outdmp->uniqueid == uniqueid) {
			dfc_rsp_data_copy(phba, (uint8_t*)buf, outdmp,
								outdmp->data);

			rc = outdmp->flag;
			list_del(&outdmp->list);
			dfc_cmd_data_free(phba, outdmp);
			break;

		}
	}
	return rc;
}
static struct bin_attribute sysfs_ctpass_attr = {
	.attr = {
		.name = "ctpass",
		.mode = S_IRUSR | S_IWUSR,
		.owner = THIS_MODULE,
	},
	.size = 0,
	.read = sysfs_ctpass_read,
	.write = sysfs_ctpass_write,
};


static ssize_t
sysfs_sendrnid_write(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
						struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	struct lpfc_sli *psli = &phba->sli;
	struct lpfc_sli_ring *pring = &psli->ring[LPFC_ELS_RING];
	struct lpfc_name idn;
	struct lpfc_iocbq *cmdiocbq = NULL, *rspiocbq = NULL;
	RNID *prsp;
	uint32_t *pcmd, *psta;
	IOCB_t *rsp;
	void *context2;
	unsigned long iflag;
	int rtnbfrsiz, i, rc = 0;
	struct lpfc_nodelist *pndl;
	uint32_t wwntype;
	struct rnidrsp *rspbuf;

	rspbuf = kmalloc(sizeof (rspbuf), GFP_KERNEL);
	if (!rspbuf)
		return -ENOMEM;

	memcpy(&idn, buf, sizeof(struct lpfc_name));
	sscanf(buf + sizeof(struct lpfc_name), "%d", &wwntype);

	spin_lock_irqsave(phba->host->host_lock, iflag); /* remove */

	if (wwntype)
		pndl = lpfc_findnode_wwpn(phba, NLP_SEARCH_MAPPED |
			 NLP_SEARCH_UNMAPPED, &idn);
	else
		pndl = lpfc_findnode_wwnn(phba, NLP_SEARCH_MAPPED | 
			NLP_SEARCH_UNMAPPED, &idn);

	if (!pndl) {
		rc = -ENODEV;
		goto sendrnid_exit;
	}

	if ((pndl->nlp_flag & NLP_ELS_SND_MASK) == NLP_RNID_SND) {
		rc = -EACCES;
		goto sendrnid_exit;
	}
	
	cmdiocbq = lpfc_prep_els_iocb(phba, 1, 2 * sizeof (uint32_t), 0, pndl,
								ELS_CMD_RNID);
	if (!cmdiocbq) {
		rc = -ENOMEM;
		goto sendrnid_exit;
	}

	/*********************************************************************/
	/*  context2 is used by prep/free to locate cmd and rsp buffers,     */
	/*  but context2 is also used by iocb_wait to hold a rspiocb ptr, so */
	/*  the rsp iocbq can be returned from the completion routine for    */
	/*  iocb_wait, so, save the prep/free value locally ... it will be   */
	/*  restored after returning from iocb_wait.                         */
	/*********************************************************************/
	context2 = cmdiocbq->context2;	/* needed to use lpfc_els_free_iocb */

	rspiocbq = mempool_alloc(phba->iocb_mem_pool, GFP_ATOMIC);
	if (!rspiocbq) {
		rc = -ENOMEM;
		goto sendrnid_freecmdiocbq;
	}
	memset(rspiocbq, 0, sizeof (struct lpfc_iocbq));
	rsp = &rspiocbq->iocb;

	pcmd = ((struct lpfc_dmabuf *) cmdiocbq->context2)->virt;
	*pcmd++ = ELS_CMD_RNID;
	memset(pcmd, 0, sizeof (RNID));	/* fill in RNID payload */
	((RNID *)pcmd)->Format = RNID_TOPOLOGY_DISC;
	cmdiocbq->context1 = NULL;
	cmdiocbq->context2 = NULL;
	cmdiocbq->iocb_flag |= LPFC_IO_LIBDFC;

	for (rc = -1, i = 0; i < 4 && rc != IOCB_SUCCESS; i++) {
		pndl->nlp_flag |= NLP_RNID_SND;
		spin_unlock_irqrestore(phba->host->host_lock, iflag);
		rc = lpfc_sli_issue_iocb_wait(phba, pring, cmdiocbq,
					      SLI_IOCB_USE_TXQ, rspiocbq,
					      (phba->fc_ratov * 2) +
					      LPFC_DRVR_TIMEOUT);
		spin_lock_irqsave(phba->host->host_lock, iflag);
		pndl->nlp_flag &= ~NLP_RNID_SND;
		cmdiocbq->context2 = context2;
		if (rc == IOCB_ERROR) {
			rc = -EACCES;
			goto sendrnid_freerspiocbq;
		}
	}

	if (rc != IOCB_SUCCESS) {
		rc = -EACCES;
		goto sendrnid_freerspiocbq;
	}

	if (rsp->ulpStatus) {
		rc = -EACCES;
	} else {
		struct lpfc_dmabuf *buf_ptr1, *buf_ptr;
		buf_ptr1 = (struct lpfc_dmabuf *)cmdiocbq->context2;
		buf_ptr = list_entry(buf_ptr1->list.next, struct lpfc_dmabuf,
									list);
		psta = (uint32_t*)buf_ptr->virt;
		if (*psta++ != ELS_CMD_ACC) {
			rc = -EFAULT;
			goto sendrnid_freerspiocbq;
		}
		prsp = (RNID*)psta;	/*  then rnid response data */
		rtnbfrsiz = prsp->CommonLen + prsp->SpecificLen;
		if (rtnbfrsiz > PAGE_SIZE) {
			rc = -EFAULT;
			goto sendrnid_freerspiocbq;
		}
		rspbuf->buf = kmalloc(rtnbfrsiz, GFP_ATOMIC);
		if (!rspbuf->buf) {
			rc = -ENOMEM;
			goto sendrnid_freerspiocbq;
		}
			
		memcpy(rspbuf->buf, prsp, rtnbfrsiz);
		rspbuf->data = rtnbfrsiz;
		rspbuf->uniqueid = current_thread_info()->task->pid;
		list_add(&rspbuf->list, &phba->rnidrspbuflist);
		rc = rtnbfrsiz;
		goto sendrnid_exit;
	}
sendrnid_freerspiocbq:
	mempool_free(rspiocbq, phba->iocb_mem_pool);
sendrnid_freecmdiocbq:
	lpfc_els_free_iocb(phba, cmdiocbq);
sendrnid_exit:
	spin_unlock_irqrestore(phba->host->host_lock, iflag);
	return rc;
}

static ssize_t
sysfs_sendrnid_read(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	int rc = -EIO, uniqueid;
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
						struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	struct rnidrsp *outdmp = NULL, *tmpoutdmp;
	
	uniqueid = current_thread_info()->task->pid;

	list_for_each_entry_safe(outdmp, tmpoutdmp, &phba->rnidrspbuflist,
									list){
		if (outdmp->uniqueid == uniqueid) {
			memcpy(buf, outdmp->buf, outdmp->data);
			rc = outdmp->data;
			kfree(outdmp->buf);
			list_del(&outdmp->list);
			kfree(outdmp);
			break;

		}
	}

	return rc;

}


static struct bin_attribute sysfs_sendrnid_attr = {
	.attr = {
		.name = "sendrnid",
		.mode = S_IWUSR,
		.owner = THIS_MODULE,
	},
	.size = 0,
	.write = sysfs_sendrnid_write,
	.read = sysfs_sendrnid_read,
};

static ssize_t
sysfs_slimem_write(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	unsigned long iflag;
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
						struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	size_t slim_size;

	if (phba->sli.sliinit.sli_flag & LPFC_SLI2_ACTIVE)
		slim_size = phba->slim_size;
	else
		slim_size = 4096;

	if ((count + off) > slim_size)
		return -ERANGE;

	if (off % 4 ||  count % 4 || (unsigned long)buf % 4)
		return -EINVAL;

	spin_lock_irqsave(host->host_lock, iflag);

	if (!(phba->fc_flag & FC_OFFLINE_MODE)) {
		if (off >= 256 && off + count <= (256 + 128)) {
			spin_unlock_irqrestore(phba->host->host_lock, iflag);
			return -EPERM;
		}
	}

	if (phba->sli.sliinit.sli_flag & LPFC_SLI2_ACTIVE)
		lpfc_sli_pcimem_bcopy((uint32_t*)buf,
			(uint32_t *)((uint8_t *) phba->slim2p.virt+off), count);
	else
		lpfc_memcpy_to_slim((uint8_t *) phba->slim_memmap_p + off,
				    (void *)buf, count);
	
	spin_unlock_irqrestore(host->host_lock, iflag);

	return count;
}

static ssize_t
sysfs_slimem_read(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	unsigned long iflag;
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
						struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];
	size_t slim_size;

	if (phba->sli.sliinit.sli_flag & LPFC_SLI2_ACTIVE)
		slim_size = phba->slim_size;
	else
		slim_size = 4096;

	if (off > slim_size)
		return -ERANGE;

	if (off % 4 ||  count % 4 || (unsigned long)buf % 4)
		return -EINVAL;

	if ((count + off) > slim_size)
		count = slim_size - off;

	spin_lock_irqsave(phba->host->host_lock, iflag);

	if (phba->sli.sliinit.sli_flag & LPFC_SLI2_ACTIVE)
		lpfc_sli_pcimem_bcopy((uint32_t *)((uint8_t *) phba->slim2p.virt
					   + off), (uint32_t *)buf, count);
	else
		lpfc_memcpy_from_slim(buf, (uint8_t *)phba->slim_memmap_p + off,
									count);

	spin_unlock_irqrestore(phba->host->host_lock, iflag);

	return count;
}

static struct bin_attribute sysfs_slimem_attr = {
	.attr = {
		.name = "slimem",
		.mode = S_IRUSR | S_IWUSR,
		.owner = THIS_MODULE,
	},
	.size = 4096,
	.read = sysfs_slimem_read,
	.write = sysfs_slimem_write,
};


static ssize_t
sysfs_ctlreg_write(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	unsigned long iflag;
	size_t buf_off;
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
					     struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];

	if ((off + count) > 255)
		return -ERANGE;

	if (off % 4 || count % 4 || (unsigned long)buf % 4)
		return -EINVAL;

	spin_lock_irqsave(phba->host->host_lock, iflag);

	if (!(phba->fc_flag & FC_OFFLINE_MODE)) {
		spin_unlock_irqrestore(phba->host->host_lock, iflag);
		return -EPERM;
	}

	for (buf_off = 0; buf_off < count; buf_off += sizeof(uint32_t))
		writel(*((uint32_t *)(buf + buf_off)),
		       (uint8_t *)phba->ctrl_regs_memmap_p + off + buf_off);

	spin_unlock_irqrestore(phba->host->host_lock, iflag);

	return count;
}

static ssize_t
sysfs_ctlreg_read(struct kobject *kobj, char *buf, loff_t off, size_t count)
{
	unsigned long iflag;
	size_t buf_off;
	uint32_t * tmp_ptr;
	struct Scsi_Host *host = class_to_shost(container_of(kobj,
					     struct class_device, kobj));
	struct lpfc_hba *phba = (struct lpfc_hba*)host->hostdata[0];

	if ((off + count) > 256)
		return -ERANGE;

	if (off % 4 || count % 4 || (unsigned long)buf % 4)
		return -EINVAL;

	spin_lock_irqsave(phba->host->host_lock, iflag);

	for (buf_off = 0; buf_off < count; buf_off += sizeof(uint32_t)) {
		tmp_ptr = (uint32_t *)(buf + buf_off);
		*tmp_ptr = readl((uint8_t *)(phba->ctrl_regs_memmap_p
					     + off + buf_off));
	}

	spin_unlock_irqrestore(phba->host->host_lock, iflag);

	return count;
}

static struct bin_attribute sysfs_ctlreg_attr = {
	.attr = {
		.name = "ctlreg",
		.mode = S_IRUSR | S_IWUSR,
		.owner = THIS_MODULE,
	},
	.size = 256,
	.read = sysfs_ctlreg_read,
	.write = sysfs_ctlreg_write,
};

static void
lpfc_get_port_id(struct scsi_device *sdev)
{
	struct lpfc_target *target = sdev->hostdata;
	if (sdev->host->transportt && target->pnode)
		fc_port_id(sdev) = target->pnode->nlp_DID;
}
static void
lpfc_get_node_name(struct scsi_device *sdev)
{
	struct lpfc_target *target = sdev->hostdata;
	uint64_t node_name = 0;
	if (sdev->host->transportt && target->pnode)
		memcpy(&node_name, &target->pnode->nlp_nodename,
						sizeof(struct lpfc_name));
	fc_node_name(sdev) = be64_to_cpu(node_name);
}
static void
lpfc_get_port_name(struct scsi_device *sdev)
{
	struct lpfc_target *target = sdev->hostdata;
	uint64_t port_name = 0;
	if (sdev->host->transportt && target->pnode)
		memcpy(&port_name, &target->pnode->nlp_portname,
						sizeof(struct lpfc_name));
	fc_port_name(sdev) = be64_to_cpu(port_name);
}
static struct fc_function_template lpfc_transport_functions = {
	.get_port_id  = lpfc_get_port_id,
	.show_port_id = 1,
	.get_node_name = lpfc_get_node_name,
	.show_node_name = 1,
	.get_port_name = lpfc_get_port_name,
	.show_port_name = 1,
};
#endif

static int
lpfc_proc_info(struct Scsi_Host *host,
	       char *buf, char **start, off_t offset, int count, int rw)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)host->hostdata[0];
	struct pci_dev *pdev = phba->pcidev;
	lpfc_vpd_t *vp = &phba->vpd;
	char fwrev[32];
	struct lpfc_nodelist *ndlp;
	char hdw[9];
	int len = 0;
	
	/* Sufficient bytes to hold a port or node name. */
	uint8_t name[sizeof (struct lpfc_name)];

	/* If rw = 0, then read info
	 * If rw = 1, then write info (NYI)
	 */
	if (rw)
		return -EINVAL;
	len += snprintf(buf, PAGE_SIZE, LPFC_MODULE_DESC "\n");

	len += snprintf(buf + len, PAGE_SIZE-len, "%s\n",
			lpfc_info(phba->host));

	len += snprintf(buf + len, PAGE_SIZE-len, "SerialNum: %s\n",
			phba->SerialNumber);

	lpfc_decode_firmware_rev(phba, fwrev, 1);
	len += snprintf(buf + len, PAGE_SIZE-len, "Firmware Version: %s\n",
			fwrev);

	len += snprintf(buf + len, PAGE_SIZE-len, "Hdw: ");
	/* Convert JEDEC ID to ascii for hardware version */
	lpfc_jedec_to_ascii(vp->rev.biuRev, hdw);
	len += snprintf(buf + len, PAGE_SIZE-len, hdw);

	len += snprintf(buf + len, PAGE_SIZE-len, "\nVendorId: 0x%x\n",
		       ((((uint32_t) pdev->device) << 16) |
			(uint32_t) (pdev->vendor)));

	/* A Fibre Channel node or port name is 8 octets long and delimited by
	 * colons.
	 */
	len += snprintf(buf + len, PAGE_SIZE-len, "Portname: ");
	memcpy (&name[0], &phba->fc_portname, sizeof (struct lpfc_name));
	len += snprintf(buf + len, PAGE_SIZE-len,
			"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
			name[0], name[1], name[2], name[3], name[4], name[5],
			name[6], name[7]);

	len += snprintf(buf + len, PAGE_SIZE-len, "   Nodename: ");
	memcpy (&name[0], &phba->fc_nodename, sizeof (struct lpfc_name));
	len += snprintf(buf + len, PAGE_SIZE-len,
			"%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x",
			name[0], name[1], name[2], name[3], name[4], name[5],
			name[6], name[7]);

	switch (phba->hba_state) {
	case LPFC_INIT_START:
	case LPFC_INIT_MBX_CMDS:
	case LPFC_LINK_DOWN:
		len += snprintf(buf + len, PAGE_SIZE-len, "\n\nLink Down\n");
		break;
	case LPFC_LINK_UP:
	case LPFC_LOCAL_CFG_LINK:
		len += snprintf(buf + len, PAGE_SIZE-len, "\n\nLink Up\n");
		break;
	case LPFC_FLOGI:
	case LPFC_FABRIC_CFG_LINK:
	case LPFC_NS_REG:
	case LPFC_NS_QRY:
	case LPFC_BUILD_DISC_LIST:
	case LPFC_DISC_AUTH:
	case LPFC_CLEAR_LA:
		len += snprintf(buf + len, PAGE_SIZE-len,
				"\n\nLink Up - Discovery\n");
		break;
	case LPFC_HBA_READY:
		len += snprintf(buf + len, PAGE_SIZE-len,
				"\n\nLink Up - Ready:\n");
		len += snprintf(buf + len, PAGE_SIZE-len, "   PortID 0x%x\n",
				phba->fc_myDID);
		if (phba->fc_topology == TOPOLOGY_LOOP) {
			if (phba->fc_flag & FC_PUBLIC_LOOP)
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Public Loop\n");
			else
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Private Loop\n");
		} else {
			if (phba->fc_flag & FC_FABRIC)
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Fabric\n");
			else
				len += snprintf(buf + len, PAGE_SIZE-len,
						"   Point-2-Point\n");
		}
		if (phba->fc_linkspeed == LA_2GHZ_LINK)
			len += snprintf(buf + len, PAGE_SIZE-len,
					"   Current speed 2G\n");
		else
			len += snprintf(buf + len, PAGE_SIZE-len,
					"   Current speed 1G\n\n");

		/* Loop through the list of mapped nodes and dump the known node
		   information. */
		list_for_each_entry(ndlp, &phba->fc_nlpmap_list, nlp_listp) {
			if (ndlp->nlp_state == NLP_STE_MAPPED_NODE){
				len += snprintf(buf + len, PAGE_SIZE -len,
						"lpfc%dt%02x DID %06x WWPN ",
						phba->brd_no,
						ndlp->nlp_sid, ndlp->nlp_DID);

				/* A Fibre Channel node or port name is 8 octets
				 * long and delimited by colons.
				 */
				memcpy (&name[0], &ndlp->nlp_portname,
					sizeof (struct lpfc_name));
				len += snprintf(buf + len, PAGE_SIZE-len,
						"%02x:%02x:%02x:%02x:%02x:%02x:"
						"%02x:%02x",
						name[0], name[1], name[2],
						name[3], name[4], name[5],
						name[6], name[7]);

				len += snprintf(buf + len, PAGE_SIZE-len,
						" WWNN ");
				memcpy (&name[0], &ndlp->nlp_nodename,
					sizeof (struct lpfc_name));
				len += snprintf(buf + len, PAGE_SIZE-len,
						"%02x:%02x:%02x:%02x:%02x:%02x:"
						"%02x:%02x\n",
						name[0], name[1], name[2],
						name[3], name[4], name[5],
						name[6], name[7]);

			}
			if (PAGE_SIZE - len < 90)
				break;
		}

		if (&ndlp->nlp_listp != &phba->fc_nlpmap_list)
			len += snprintf(buf+len, PAGE_SIZE-len, "...\n");

	}
	return (len);
}

static int
lpfc_slave_alloc(struct scsi_device *scsi_devs)
{
	struct lpfc_hba *phba;
	struct lpfc_target *target;

	/* 
	 * Store the lun pointer in the scsi_device hostdata pointer provided
	 * the driver has already discovered the target id.  
	 */
	phba = (struct lpfc_hba *) scsi_devs->host->hostdata[0];
	target = lpfc_find_target(phba, scsi_devs->id);
	if (target) {
		scsi_devs->hostdata = target;
		return 0;
	}
		
	/* 
	 * The driver does not have a target id matching that in the scsi
	 * device.  Allocate a dummy target initialized to zero so that
	 * the driver's queuecommand entry correctly fails the call
	 * forcing the midlayer to call lpfc_slave_destroy.  This code
	 * will be removed in a subsequent kernel patch.
	 */

	target = kmalloc(sizeof (struct lpfc_target), GFP_KERNEL);
	if (!target)
		return 1;

	memset(target, 0, sizeof (struct lpfc_target));
	scsi_devs->hostdata = target;
	return 0;
}

static int
lpfc_slave_configure(struct scsi_device *sdev)
{
	struct lpfc_hba *phba = (struct lpfc_hba *)sdev->host->hostdata[0];

	if (sdev->tagged_supported) {
		sdev->current_tag = 0;
		sdev->queue_depth = phba->cfg_lun_queue_depth;
	}
	return 0;
}

static void
lpfc_slave_destroy(struct scsi_device *sdev)
{
	struct lpfc_hba *phba;
	struct lpfc_target *target;

	phba = (struct lpfc_hba *) sdev->host->hostdata[0];
	target = sdev->hostdata;

	if (target) {
		if ((!target->pnode) && (target->targetFlags == 0))
			kfree(target);
		sdev->hostdata = 0;
	}
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
static struct class_device_attribute *lpfc_host_attrs[] = {
	&class_device_attr_info,
	&class_device_attr_serialnum,
	&class_device_attr_fwrev,
	&class_device_attr_hdw,
	&class_device_attr_option_rom_version,
	&class_device_attr_num_discovered_ports,
	&class_device_attr_node_name,
	&class_device_attr_port_name,
	&class_device_attr_portfcid,
	&class_device_attr_events,
	&class_device_attr_lpfc_drvr_version,
	&class_device_attr_lpfc_log_verbose,
	&class_device_attr_lpfc_lun_queue_depth,
	&class_device_attr_lpfc_nodev_tmo,
	&class_device_attr_lpfc_automap,
	&class_device_attr_lpfc_fcp_class,
	&class_device_attr_lpfc_use_adisc,
	&class_device_attr_lpfc_ack0,	
	&class_device_attr_lpfc_topology,	
	&class_device_attr_lpfc_scan_down,
	&class_device_attr_lpfc_link_speed,
	&class_device_attr_lpfc_fdmi_on,
	&class_device_attr_lpfc_fcp_bind_method,
	&class_device_attr_nport_evt_cnt,
	&class_device_attr_board_online,
	NULL,
};
#endif
static struct scsi_host_template driver_template = {
	.module			= THIS_MODULE,
	.name			= LPFC_DRIVER_NAME,
	.info			= lpfc_info,
	.queuecommand		= lpfc_queuecommand,
	.eh_abort_handler	= lpfc_abort_handler,
	.eh_device_reset_handler= lpfc_reset_lun_handler,
	.eh_bus_reset_handler	= lpfc_reset_bus_handler,
	.slave_alloc		= lpfc_slave_alloc,
	.slave_configure	= lpfc_slave_configure,
	.slave_destroy		= lpfc_slave_destroy,
	.proc_info		= lpfc_proc_info,
	.proc_name		= LPFC_DRIVER_NAME,
	.this_id		= -1,
	.sg_tablesize		= SG_ALL,
	.cmd_per_lun		= 30,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
	.shost_attrs		= lpfc_host_attrs,
#endif
	.use_clustering		= ENABLE_CLUSTERING,
};

void
lpfc_setup_slim_access(struct lpfc_hba *phba)
{
	phba->MBslimaddr = phba->slim_memmap_p;
	phba->HAregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		HA_REG_OFFSET;
	phba->HCregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		HC_REG_OFFSET;
	phba->CAregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		CA_REG_OFFSET;
	phba->HSregaddr = (uint32_t *) (phba->ctrl_regs_memmap_p) +
		HS_REG_OFFSET;
	return;
}


uint32_t
lpfc_intr_prep(struct lpfc_hba * phba)
{
	uint32_t ha_copy;

	/* Ignore all interrupts during initialization. */
	if (phba->hba_state < LPFC_LINK_DOWN)
		return (0);

	/* Read host attention register to determine interrupt source */
	ha_copy = readl(phba->HAregaddr);

	/* Clear Attention Sources, except ERATT (to preserve status) & LATT
	 *    (ha_copy & ~(HA_ERATT | HA_LATT));
	 */
	writel((ha_copy & ~(HA_LATT | HA_ERATT)), phba->HAregaddr);
	readl(phba->HAregaddr); /* flush */
	return (ha_copy);
}				/* lpfc_intr_prep */

static int
lpfc_sli_setup(struct lpfc_hba * phba)
{
	int i, totiocb = 0;
	struct lpfc_sli *psli = &phba->sli;
	LPFC_RING_INIT_t *pring;

	psli->sliinit.num_rings = MAX_CONFIGURED_RINGS;
	psli->fcp_ring = LPFC_FCP_RING;
	psli->next_ring = LPFC_FCP_NEXT_RING;
	psli->ip_ring = LPFC_IP_RING;

	for (i = 0; i < psli->sliinit.num_rings; i++) {
		pring = &psli->sliinit.ringinit[i];
		switch (i) {
		case LPFC_FCP_RING:	/* ring 0 - FCP */
			/* numCiocb and numRiocb are used in config_port */
			pring->numCiocb = SLI2_IOCB_CMD_R0_ENTRIES;
			pring->numRiocb = SLI2_IOCB_RSP_R0_ENTRIES;
			pring->numCiocb += SLI2_IOCB_CMD_R1XTRA_ENTRIES;
			pring->numRiocb += SLI2_IOCB_RSP_R1XTRA_ENTRIES;
			pring->numCiocb += SLI2_IOCB_CMD_R3XTRA_ENTRIES;
			pring->numRiocb += SLI2_IOCB_RSP_R3XTRA_ENTRIES;
			pring->iotag_ctr = 0;
			pring->iotag_max =
			    (phba->cfg_hba_queue_depth * 2);
			pring->fast_iotag = pring->iotag_max;
			pring->num_mask = 0;
			break;
		case LPFC_IP_RING:	/* ring 1 - IP */
			/* numCiocb and numRiocb are used in config_port */
			pring->numCiocb = SLI2_IOCB_CMD_R1_ENTRIES;
			pring->numRiocb = SLI2_IOCB_RSP_R1_ENTRIES;
			pring->num_mask = 0;
			break;
		case LPFC_ELS_RING:	/* ring 2 - ELS / CT */
			/* numCiocb and numRiocb are used in config_port */
			pring->numCiocb = SLI2_IOCB_CMD_R2_ENTRIES;
			pring->numRiocb = SLI2_IOCB_RSP_R2_ENTRIES;
			pring->fast_iotag = 0;
			pring->iotag_ctr = 0;
			pring->iotag_max = 4096;
			pring->num_mask = 4;
			pring->prt[0].profile = 0;	/* Mask 0 */
			pring->prt[0].rctl = FC_ELS_REQ;
			pring->prt[0].type = FC_ELS_DATA;
			pring->prt[0].lpfc_sli_rcv_unsol_event =
			    lpfc_els_unsol_event;
			pring->prt[1].profile = 0;	/* Mask 1 */
			pring->prt[1].rctl = FC_ELS_RSP;
			pring->prt[1].type = FC_ELS_DATA;
			pring->prt[1].lpfc_sli_rcv_unsol_event =
			    lpfc_els_unsol_event;
			pring->prt[2].profile = 0;	/* Mask 2 */
			/* NameServer Inquiry */
			pring->prt[2].rctl = FC_UNSOL_CTL;
			/* NameServer */
			pring->prt[2].type = FC_COMMON_TRANSPORT_ULP;
			pring->prt[2].lpfc_sli_rcv_unsol_event =
			    lpfc_ct_unsol_event;
			pring->prt[3].profile = 0;	/* Mask 3 */
			/* NameServer response */
			pring->prt[3].rctl = FC_SOL_CTL;
			/* NameServer */
			pring->prt[3].type = FC_COMMON_TRANSPORT_ULP;
			pring->prt[3].lpfc_sli_rcv_unsol_event =
			    lpfc_ct_unsol_event;
			break;
		}
		totiocb += (pring->numCiocb + pring->numRiocb);
	}
	if (totiocb > MAX_SLI2_IOCB) {
		/* Too many cmd / rsp ring entries in SLI2 SLIM */
		lpfc_printf_log(phba, KERN_ERR, LOG_INIT,
				"%d:0462 Too many cmd / rsp ring entries in "
				"SLI2 SLIM Data: x%x x%x\n",
				phba->brd_no, totiocb, MAX_SLI2_IOCB);
	}

#ifdef USE_HGP_HOST_SLIM
	psli->sliinit.sli_flag = LPFC_HGP_HOSTSLIM;
#else
	psli->sliinit.sli_flag = 0;
#endif

	return (0);
}

irqreturn_t
lpfc_intr_handler(int irq, void *dev_id, struct pt_regs * regs)
{
	struct lpfc_hba *phba;
	int intr_status;

	/*
	 * Get the driver's phba structure from the dev_id and
	 * assume the HBA is not interrupting.
	 */
	phba = (struct lpfc_hba *) dev_id;

	if (phba) {
		/* Call SLI to handle the interrupt event. */
		intr_status = lpfc_sli_intr(phba);
		if (intr_status == 0)
			return IRQ_HANDLED;
	}

	return IRQ_NONE;

} /* lpfc_intr_handler */

static int
lpfc_set_bind_type(struct lpfc_hba * phba)
{
	int bind_type = phba->cfg_fcp_bind_method;
	int ret = LPFC_BIND_WW_NN_PN;

	switch (bind_type) {
	case 1:
		phba->fcp_mapping = FCP_SEED_WWNN;
		break;

	case 2:
		phba->fcp_mapping = FCP_SEED_WWPN;
		break;

	case 3:
		phba->fcp_mapping = FCP_SEED_DID;
		ret = LPFC_BIND_DID;
		break;

	case 4:
		phba->fcp_mapping = FCP_SEED_DID;
		ret = LPFC_BIND_DID;
		break;
	}

	return (ret);
}

static void
lpfc_wakeup_event(struct lpfc_hba * phba, fcEVTHDR_t * ep)
{
	ep->e_mode &= ~E_SLEEPING_MODE;
	switch (ep->e_mask) {
	case FC_REG_LINK_EVENT:
		wake_up_interruptible(&phba->linkevtwq);
		break;
	case FC_REG_RSCN_EVENT:
		wake_up_interruptible(&phba->rscnevtwq);
		break;
	case FC_REG_CT_EVENT:
		wake_up_interruptible(&phba->ctevtwq);
		break;
	}
	return;
}

static void
lpfc_get_cfgparam(struct lpfc_hba *phba)
{
	phba->cfg_log_verbose = lpfc_log_verbose;
	phba->cfg_automap = lpfc_automap;
	phba->cfg_fcp_bind_method = lpfc_fcp_bind_method;
	phba->cfg_cr_delay = lpfc_cr_delay;
	phba->cfg_cr_count = lpfc_cr_count;
	phba->cfg_lun_queue_depth = lpfc_lun_queue_depth;
	phba->cfg_fcp_class = lpfc_fcp_class;
	phba->cfg_use_adisc = lpfc_use_adisc;
	phba->cfg_ack0 = lpfc_ack0;
	phba->cfg_topology = lpfc_topology;
	phba->cfg_scan_down = lpfc_scan_down;
	phba->cfg_nodev_tmo = lpfc_nodev_tmo;
	phba->cfg_link_speed = lpfc_link_speed;
	phba->cfg_fdmi_on = lpfc_fdmi_on;
	phba->cfg_discovery_threads = lpfc_discovery_threads;

	if (phba->cfg_discovery_threads)
		if (phba->cfg_automap == 0)
			phba->cfg_discovery_threads = LPFC_MAX_DISC_THREADS;

	switch (phba->pcidev->device) {
	case PCI_DEVICE_ID_LP101:
		phba->cfg_hba_queue_depth = LPFC_LP101_HBA_Q_DEPTH;
		break;
	case PCI_DEVICE_ID_RFLY:
	case PCI_DEVICE_ID_PFLY:
	case PCI_DEVICE_ID_JFLY:
	case PCI_DEVICE_ID_ZFLY:
	case PCI_DEVICE_ID_TFLY:
		phba->cfg_hba_queue_depth = LPFC_LC_HBA_Q_DEPTH;
		break;
	default:
		phba->cfg_hba_queue_depth = LPFC_DFT_HBA_Q_DEPTH;
	}
	return;
}

int
lpfc_put_event(struct lpfc_hba * phba, uint32_t evcode, uint32_t evdata0,
	       void * evdata1, uint32_t evdata2, uint32_t evdata3)
{
	fcEVT_t *ep;
	fcEVTHDR_t *ehp = phba->fc_evt_head;
	int found = 0;
	void *fstype = NULL;
	struct lpfc_dmabuf *mp;
	struct lpfc_sli_ct_request *ctp;
	struct lpfc_hba_event *rec;
	uint32_t evtype;

	switch (evcode) {
		case HBA_EVENT_RSCN:
			evtype = FC_REG_RSCN_EVENT;
			break;
		case HBA_EVENT_LINK_DOWN:
		case HBA_EVENT_LINK_UP:
			evtype = FC_REG_LINK_EVENT;
			break;
		default:
			evtype = FC_REG_CT_EVENT;
	}

	if (evtype == FC_REG_RSCN_EVENT || evtype == FC_REG_LINK_EVENT) {
		rec = &phba->hbaevt[phba->hba_event_put];
		rec->fc_eventcode = evcode;
		rec->fc_evdata1 = evdata0;
		rec->fc_evdata2 = (uint32_t)(unsigned long)evdata1;
		rec->fc_evdata3 = evdata2;
		rec->fc_evdata4 = evdata3;

		phba->hba_event_put++;
		if (phba->hba_event_put >= MAX_HBAEVT)
			phba->hba_event_put = 0;

		if (phba->hba_event_put == phba->hba_event_get) {
			phba->hba_event_missed++;
			phba->hba_event_get++;
			if (phba->hba_event_get >= MAX_HBAEVT)
				phba->hba_event_get = 0;
		}
	}

	if (evtype == FC_REG_CT_EVENT) {
		mp = (struct lpfc_dmabuf *) evdata1;
		ctp = (struct lpfc_sli_ct_request *) mp->virt;
		fstype = (void *)(ulong) (ctp->FsType);
	}

	while (ehp && ((ehp->e_mask != evtype) || (ehp->e_type != fstype)))
		ehp = (fcEVTHDR_t *) ehp->e_next_header;

	if (!ehp)
		return (0);

	ep = ehp->e_head;

	while (ep && !(found)) {
		if (ep->evt_sleep) {
			switch (evtype) {
			case FC_REG_CT_EVENT:
				if ((ep->evt_type ==
				     (void *)(ulong) FC_FSTYPE_ALL)
				    || (ep->evt_type == fstype)) {
					found++;
					ep->evt_data0 = evdata0; /* tag */
					ep->evt_data1 = evdata1; /* buffer
								    ptr */
					ep->evt_data2 = evdata2; /* count */
					ep->evt_sleep = 0;
					if (ehp->e_mode & E_SLEEPING_MODE) {
						ehp->e_flag |=
						    E_GET_EVENT_ACTIVE;
						lpfc_wakeup_event(phba, ehp);
					}
					/* For FC_REG_CT_EVENT just give it to
					   first one found */
				}
				break;
			default:
				found++;
				ep->evt_data0 = evdata0;
				ep->evt_data1 = evdata1;
				ep->evt_data2 = evdata2;
				ep->evt_sleep = 0;
				if ((ehp->e_mode & E_SLEEPING_MODE)
				    && !(ehp->e_flag & E_GET_EVENT_ACTIVE)) {
					ehp->e_flag |= E_GET_EVENT_ACTIVE;
					lpfc_wakeup_event(phba, ehp);
				}
				/* For all other events, give it to every one
				   waiting */
				break;
			}
		}
		ep = ep->evt_next;
	}
	if (evtype == FC_REG_LINK_EVENT)
		phba->nport_event_cnt++;

	return (found);
}

static void
lpfc_consistent_bind_setup(struct lpfc_hba * phba)
{
	INIT_LIST_HEAD(&phba->fc_nlpbind_list);
	phba->fc_bind_cnt = 0;
}

static uint8_t
lpfc_get_brd_no(struct lpfc_hba * phba)
{
	uint8_t    brd, found = 1;

 	brd = 0;
	while(found) {
		phba = NULL;
		found = 0;
		list_for_each_entry(phba, &lpfc_hba_list, hba_list) {
			if (phba->brd_no == brd) {
				found = 1;
				brd++;
				break;
			}
		}
	}
	return (brd);
}	

static int __devinit
lpfc_pci_probe_one(struct pci_dev *pdev, const struct pci_device_id *pid)
{
	struct Scsi_Host *host;
	struct lpfc_hba  *phba;
	struct lpfc_sli  *psli;
	unsigned long iflag;
	unsigned long bar0map_len, bar2map_len;
	int error = -ENODEV, retval;

	if (pci_enable_device(pdev))
		goto out;
	if (pci_request_regions(pdev, LPFC_DRIVER_NAME))
		goto out_disable_device;

	/*
	 * Allocate space for adapter info structure
	 */
	phba = kmalloc(sizeof(*phba), GFP_KERNEL);
	if (!phba)
		goto out_release_regions;
	memset(phba, 0, sizeof (struct lpfc_hba));

	host = scsi_host_alloc(&driver_template, sizeof (unsigned long));
	if (!host) {
		printk (KERN_WARNING "%s: scsi_host_alloc failed.\n",
							 lpfc_drvr_name);
		error = -ENOMEM;
		goto out_kfree_phba;
	}
	host->max_id = LPFC_MAX_TARGET;
	host->max_lun = LPFC_MAX_LUN; 
	host->this_id = -1;

	phba->pcidev = pdev;
	phba->host = host;

	INIT_LIST_HEAD(&phba->ctrspbuflist);
	INIT_LIST_HEAD(&phba->rnidrspbuflist);
	INIT_LIST_HEAD(&phba->freebufList);

	/* Initialize timers used by driver */
	init_timer(&phba->fc_estabtmo);
	phba->fc_estabtmo.function = lpfc_establish_link_tmo;
	phba->fc_estabtmo.data = (unsigned long)phba;
	init_timer(&phba->fc_disctmo);
	phba->fc_disctmo.function = lpfc_disc_timeout;
	phba->fc_disctmo.data = (unsigned long)phba;
	
	init_timer(&phba->fc_fdmitmo);
	phba->fc_fdmitmo.function = lpfc_fdmi_tmo;
	phba->fc_fdmitmo.data = (unsigned long)phba;
	init_timer(&phba->els_tmofunc);
	phba->els_tmofunc.function = lpfc_els_timeout_handler;
	phba->els_tmofunc.data = (unsigned long)phba;
	psli = &phba->sli;
	init_timer(&psli->mbox_tmo);
	psli->mbox_tmo.function = lpfc_mbox_timeout;
	psli->mbox_tmo.data = (unsigned long)phba;

	/* Assign an unused board number */
 	phba->brd_no = lpfc_get_brd_no(phba);
	host->unique_id = phba->brd_no;

	lpfc_get_cfgparam(phba);

	/* Add adapter structure to list */
	list_add_tail(&phba->hba_list, &lpfc_hba_list);

	/* Initialize all internally managed lists. */
	INIT_LIST_HEAD(&phba->fc_nlpmap_list);
	INIT_LIST_HEAD(&phba->fc_nlpunmap_list);
	INIT_LIST_HEAD(&phba->fc_unused_list);
	INIT_LIST_HEAD(&phba->fc_plogi_list);
	INIT_LIST_HEAD(&phba->fc_adisc_list);
	INIT_LIST_HEAD(&phba->fc_reglogin_list);
	INIT_LIST_HEAD(&phba->fc_prli_list);
	INIT_LIST_HEAD(&phba->fc_npr_list);
	lpfc_consistent_bind_setup(phba);

	init_waitqueue_head(&phba->linkevtwq);
	init_waitqueue_head(&phba->rscnevtwq);
	init_waitqueue_head(&phba->ctevtwq);

	pci_set_master(pdev);
	retval = pci_set_mwi(pdev);
	if (retval)
		dev_printk(KERN_WARNING, &pdev->dev,
			   "Warning: pci_set_mwi returned %d\n", retval);
	
	/* Configure DMA attributes. */
	if (dma_set_mask(&phba->pcidev->dev, 0xffffffffffffffffULL) &&
	    dma_set_mask(&phba->pcidev->dev, 0xffffffffULL))
		goto out_list_del;

	/*
	 * Get the physical address of Bar0 and Bar2 and the number of bytes
	 * required by each mapping.
	 */
	phba->pci_bar0_map = pci_resource_start(phba->pcidev, 0);
	bar0map_len        = pci_resource_len(phba->pcidev, 0);
	
	phba->pci_bar2_map = pci_resource_start(phba->pcidev, 2);
	bar2map_len        = pci_resource_len(phba->pcidev, 2);

	/* Map HBA SLIM and Control Registers to a kernel virtual address. */
	phba->slim_memmap_p      = ioremap(phba->pci_bar0_map, bar0map_len);
	phba->ctrl_regs_memmap_p = ioremap(phba->pci_bar2_map, bar2map_len);

	/* Setup SLI2 interface */
	if (!phba->slim2p.virt) {
		/*
		 * Allocate memory for SLI-2 structures
		 */
		phba->slim2p.virt = dma_alloc_coherent(&phba->pcidev->dev,
						       sizeof (SLI2_SLIM_t),
						       &(phba->slim2p.phys),
						       GFP_ATOMIC);
		if (!phba->slim2p.virt)
			goto out_iounmap;

		/* The SLIM2 size is stored in the next field */
		phba->slim_size = sizeof (SLI2_SLIM_t);
		memset(phba->slim2p.virt, 0, sizeof (SLI2_SLIM_t));
	}

	lpfc_sli_setup(phba);	/* Setup SLI Layer to run over lpfc HBAs */
	lpfc_sli_queue_setup(phba);	/* Initialize the SLI Layer */
	
	error = lpfc_mem_alloc(phba);
	if (error)
		goto out_dec_nhbas;

	lpfc_set_bind_type(phba);

	/* Initialize HBA structure */
	phba->fc_edtov = FF_DEF_EDTOV;
	phba->fc_ratov = FF_DEF_RATOV;
	phba->fc_altov = FF_DEF_ALTOV;
	phba->fc_arbtov = FF_DEF_ARBTOV;

	/* Set the FARP and XRI timeout values now since they depend on
	   fc_ratov. */
	phba->fc_ipfarp_timeout = (3 * phba->fc_ratov);
	phba->fc_ipxri_timeout = (3 * phba->fc_ratov);

	INIT_LIST_HEAD(&phba->dpc_disc);
	init_completion(&phba->dpc_startup);
	init_completion(&phba->dpc_exiting);

	/*
	* Startup the kernel thread for this host adapter
	*/
	phba->dpc_kill = 0;
	phba->dpc_pid = kernel_thread(lpfc_do_dpc, phba, 0);
	if (phba->dpc_pid < 0) {
		error = phba->dpc_pid;
		goto out_free_mem;
	}
	wait_for_completion(&phba->dpc_startup);

	/* Call SLI to initialize the HBA. */
	error = lpfc_sli_hba_setup(phba);
	if (error)
		goto out_hba_down;

	/* We can rely on a queue depth attribute only after SLI HBA setup */
	host->can_queue = phba->cfg_hba_queue_depth - 10;

	/*
	 * Starting with 2.4.0 kernel, Linux can support commands longer
	 * than 12 bytes. However, scsi_register() always sets it to 12.
	 * For it to be useful to the midlayer, we have to set it here.
	 */
	host->max_cmd_len = 16;

	/*
	 * Queue depths per lun
	 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
	host->transportt = lpfc_transport_template;
#endif
	host->hostdata[0] = (unsigned long)phba;
	pci_set_drvdata(pdev, host);
	error = scsi_add_host(host, &pdev->dev);
	if (error)
		goto out_put_host;

	//sysfs_create_bin_file(&host->shost_classdev.kobj, &sysfs_stats_attr);
	//sysfs_create_bin_file(&host->shost_classdev.kobj, &sysfs_ctpass_attr);
	//sysfs_create_bin_file(&host->shost_classdev.kobj,
	//						&sysfs_sendrnid_attr);


	if (phba->sli.sliinit.sli_flag & LPFC_SLI2_ACTIVE)
		sysfs_slimem_attr.size = phba->slim_size;
	else
		sysfs_slimem_attr.size = 4096;

	//sysfs_create_bin_file(&host->shost_classdev.kobj, &sysfs_slimem_attr);
	//sysfs_create_bin_file(&host->shost_classdev.kobj, &sysfs_ctlreg_attr);
	scsi_scan_host(host);
	return 0;

out_put_host:
	scsi_host_put(host);
out_hba_down:
	lpfc_sli_hba_down(phba);

	/* Stop any timers that were started during this attach. */
	spin_lock_irqsave(phba->host->host_lock, iflag);
	lpfc_stop_timer(phba);
	spin_unlock_irqrestore(phba->host->host_lock, iflag);

	/* Kill the kernel thread for this host */
	if (phba->dpc_pid >= 0) {
		phba->dpc_kill = 1;
		wmb();
		kill_proc(phba->dpc_pid, SIGHUP, 1);
		wait_for_completion(&phba->dpc_exiting);
	}

	free_irq(phba->pcidev->irq, phba);
out_free_mem:
	lpfc_mem_free(phba);
out_dec_nhbas:
	if (phba->slim2p.virt) {
		dma_free_coherent(&pdev->dev,
				  phba->slim_size,
				  phba->slim2p.virt,
				  phba->slim2p.phys);
	}
out_iounmap:
	iounmap(phba->ctrl_regs_memmap_p);
	iounmap(phba->slim_memmap_p);
out_list_del:
	list_del_init(&phba->hba_list);
out_kfree_phba:
	kfree(phba);
out_release_regions:
	pci_release_regions(pdev);
out_disable_device:
	pci_disable_device(pdev);
out:
	return error;
}

int
lpfc_stop_timer(struct lpfc_hba * phba)
{
	struct lpfc_sli *psli = &phba->sli;

	/* Instead of a timer, this has been converted to a
	 * deferred procedding list.
	 */
	while (!list_empty(&phba->freebufList)) {
		struct lpfc_dmabuf *mp;

		mp = (struct lpfc_dmabuf *)(phba->freebufList.next);
		if (mp) {
			lpfc_mbuf_free(phba, mp->virt, mp->phys);
			list_del(&mp->list);
			kfree((void *)mp);
		}
	}

	del_timer_sync(&phba->fc_estabtmo);
	del_timer_sync(&phba->fc_disctmo);
	del_timer_sync(&phba->fc_fdmitmo);
	del_timer_sync(&phba->els_tmofunc);
	psli = &phba->sli;
	del_timer_sync(&psli->mbox_tmo);
	return(1);
}

static void __devexit
lpfc_pci_remove_one(struct pci_dev *pdev)
{
	struct Scsi_Host *host = pci_get_drvdata(pdev);
	struct lpfc_hba  *phba = (struct lpfc_hba *)host->hostdata[0];
	unsigned long iflag;

	//sysfs_remove_bin_file(&host->shost_classdev.kobj, &sysfs_ctlreg_attr);
	//sysfs_remove_bin_file(&host->shost_classdev.kobj, &sysfs_slimem_attr);
	//sysfs_remove_bin_file(&host->shost_classdev.kobj,
	//						&sysfs_sendrnid_attr);
	//sysfs_remove_bin_file(&host->shost_gendev.kobj, &sysfs_ctpass_attr);
	//sysfs_remove_bin_file(&host->shost_gendev.kobj, &sysfs_stats_attr);
	

/* FIX ME */

	/* In case we are offline or link is down */
	/*scsi_unblock_requests(phba->host);*/

	scsi_remove_host(phba->host);
	list_del(&phba->hba_list);

	/* detach the board */

	/* Kill the kernel thread for this host */
	if (phba->dpc_pid >= 0) {
		phba->dpc_kill = 1;
		wmb();
		kill_proc(phba->dpc_pid, SIGHUP, 1);
		wait_for_completion(&phba->dpc_exiting);
	}

	/*
	 * Bring down the SLI Layer. This step disable all interrupts,
	 * clears the rings, discards all mailbox commands, and resets
	 * the HBA.
	 */
	lpfc_sli_hba_down(phba);	

	/* Release the irq reservation */
	free_irq(phba->pcidev->irq, phba);

	lpfc_cleanup(phba, 0);
	lpfc_scsi_free(phba);

	spin_lock_irqsave(phba->host->host_lock, iflag);
	lpfc_stop_timer(phba);
	spin_unlock_irqrestore(phba->host->host_lock, iflag);

	lpfc_mem_free(phba);

	/* Free resources associated with SLI2 interface */
	if (phba->slim2p.virt) {
		dma_free_coherent(&pdev->dev,
				  phba->slim_size,
				  phba->slim2p.virt,
				  phba->slim2p.phys);
	}

	/* unmap adapter SLIM and Control Registers */
	iounmap(phba->ctrl_regs_memmap_p);
	iounmap(phba->slim_memmap_p);
	
	pci_release_regions(phba->pcidev);
	pci_disable_device(phba->pcidev);

	scsi_host_put(phba->host);
	kfree(phba);

	pci_set_drvdata(pdev, 0);
}

static struct pci_device_id lpfc_id_table[] = {
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_VIPER,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_THOR,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_PEGASUS,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_CENTAUR,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_DRAGONFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_SUPERFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_RFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_PFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_HELIOS,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_JFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_ZEPHYR,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_ZFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{PCI_VENDOR_ID_EMULEX, PCI_DEVICE_ID_TFLY,
		PCI_ANY_ID, PCI_ANY_ID, },
	{ 0 }
};
MODULE_DEVICE_TABLE(pci, lpfc_id_table);


static struct pci_driver lpfc_driver = {
	.name		= LPFC_DRIVER_NAME,
	.id_table	= lpfc_id_table,
	.probe		= lpfc_pci_probe_one,
	.remove		= __devexit_p(lpfc_pci_remove_one),
};

static int __init
lpfc_init(void)
{
	int rc;

	printk(LPFC_MODULE_DESC "\n");


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
	lpfc_transport_template =
		fc_attach_transport(&lpfc_transport_functions);
	if (!lpfc_transport_template)
		return -ENODEV;
#endif
	rc = pci_module_init(&lpfc_driver);
	return rc;

}

static void __exit
lpfc_exit(void)
{
	pci_unregister_driver(&lpfc_driver);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6)
	fc_release_transport(lpfc_transport_template);
#endif
}
module_init(lpfc_init);
module_exit(lpfc_exit);
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION(LPFC_MODULE_DESC);
MODULE_AUTHOR("Emulex Corporation - tech.support@emulex.com");
MODULE_VERSION("0:" LPFC_DRIVER_VERSION);
