/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_scsi.h 1.58 2004/09/14 14:57:03EDT sf_support Exp  $
 */

#ifndef _H_LPFC_SCSI
#define _H_LPFC_SCSI

#include "lpfc_disc.h"
#include "lpfc_mem.h"
#include "lpfc_sli.h"

/*
 * SCSI node structure for each open Fibre Channel node
 * used by scsi transport.
 */

struct lpfc_hba;

struct lpfc_target {
	struct lpfc_nodelist *pnode;	/* Pointer to the node structure. */

	uint16_t max_lun;	/* max lun supported */
	uint16_t scsi_id;	/* SCSI ID of this device */
	uint16_t rpi;
	uint16_t targetFlags;
#define FC_NODEV_TMO        0x1	/* nodev-tmo tmr started and expired */
#define FC_FCP2_RECOVERY    0x2	/* set FCP2 Recovery for commands */
#define FC_NPR_ACTIVE       0x10	/* NPort Recovery active */

	uint32_t qcmdcnt;
	uint32_t iodonecnt;
	uint32_t errorcnt;

	union {
		uint32_t dev_did;	/* SCSI did */
	} un;

};

#define LPFC_SCSI_BUF_SZ        1024  /* used for driver generated scsi cmds */
#define LPFC_SCSI_PAGE_BUF_SZ   4096  /* used for driver RPTLUN cmds */
#define LPFC_INQSN_SZ           64    /* Max size of Inquiry serial number */

struct lpfc_scsi_buf {
	uint32_t scsitmo;	/* IN */
	uint32_t timeout;	/* IN */
	struct lpfc_hba *scsi_hba;	/* IN */

	struct scsi_cmnd *pCmd;	/* IN */

	uint16_t flags;		/* flags for this cmd */
#define DATA_MAPPED     0x0001	/* data buffer has been D_MAPed */
#define FCBUF_ABTS      0x0002	/* ABTS has been sent for this cmd */
#define FCBUF_ABTS2     0x0004	/* ABTS has been sent twice */
#define FCBUF_INTERNAL  0x0008	/* Internal generated driver command */
#define LPFC_SCSI_ERR    0x0010
	uint16_t IOxri;		/* From IOCB Word 6- ulpContext */
	uint16_t status;	/* From IOCB Word 7- ulpStatus */
	uint32_t result;	/* From IOCB Word 4. */

	int        datadir;	/* Data direction as requested in the scsi
				   command. */
	dma_addr_t nonsg_phys;	/* Non scatter-gather physical address. */

	struct lpfc_target *target;

	/* dma_ext has both virt, phys to dma-able buffer
	 * which contains fcp_cmd, fcp_rsp and scatter gather list fro upto
	 * 68 (LPFC_SCSI_BPL_SIZE) BDE entries,
	 * xfer length, cdb, data direction....
	 */
	struct lpfc_dmabuf dma_ext;
	struct fcp_cmnd *fcp_cmnd;
	struct fcp_rsp *fcp_rsp;
	struct ulp_bde64 *fcp_bpl;

	/* cur_iocbq has phys of the dma-able buffer.
	 * Iotag is in here
	 */
	struct lpfc_iocbq cur_iocbq;

	void (*cmd_cmpl) (struct lpfc_hba *, struct lpfc_scsi_buf *);	/* IN */
};

#define LPFC_SCSI_INITIAL_BPL_SIZE  4	/* Number of scsi buf BDEs in fcp_bpl */

#define LPFC_CMD_STATUS_ABORTED -1

#define LPFC_SCSI_DMA_EXT_SIZE 264
#define LPFC_BPL_SIZE          1024

#define MDAC_DIRECT_CMD                  0x22

#endif				/* _H_LPFC_SCSI */
